#!/usr/bin/perl

use strict;
use warnings;
use open ':encoding(utf8)';
binmode STDIN,':encoding(utf8)';
binmode STDOUT,':encoding(utf8)';


#Set these appropriately before running
my $FIND_CONLL_DOCS_SCRIPT="./extract_doc_names_from_conll.sh";
my $CLEAR_PARSER_LIB="clearparser-lib";
my $CLEAR_PARSER_JAR="clearparser-0.32.jar";

my $CLASS_PATH_DELIMITER=":"; # colon for *nix, semicolon for windows

my $JAVA_CP="$CLEAR_PARSER_LIB/lib/*".$CLASS_PATH_DELIMITER."$CLEAR_PARSER_JAR";
my $CLEAR_HEAD_RULES="$CLEAR_PARSER_LIB/config/headrule_en_ontonotes.txt";
my $CLEAR_LEMMA_DICT="$CLEAR_PARSER_LIB/model/en_dict-1.0.jar";

die("Not enough arguments. Check the code for help") if(scalar(@ARGV)<4);

my $conllSetRoot=shift; #This should be something like ontonotes5/conll-2012/v1/data/train/data/english/
my $ontoNotesRoot=shift; #This should be something like ontonotes5/conll-2012-train-v0/data/files/data/english/
my $suffix=shift; #something like v4_auto_conll
my $outRoot=shift; #The root of the output

my @p=`$FIND_CONLL_DOCS_SCRIPT $conllSetRoot $suffix`;

my $fileCount=0;
foreach my $prefix (@p){
    chomp $prefix;
#    next if($prefix ne '/annotations/tc/ch/00/ch_0011');
    print "$prefix\n";
    $fileCount++;
    my $parse_file=$ontoNotesRoot . $prefix . ".parse";
    die ("Parse tree file $parse_file does not exist!") if(! -e $parse_file);
    my $outFolder;
    my $filePrefix;
    if($prefix=~/^.*(\/(?:bc|bn|mz|nw|pt|tc|wb)\/.*\/)(.*)$/){
	$outFolder=$outRoot ."/$1";
	$filePrefix=$2;
    } else {
	die("Failed to parse dir structure in $prefix");
	# #DEBUG FOR TC:
	# $outFolder=$outRoot ."/tc/$1";
	# $filePrefix=$1 if($prefix=~/.*\/(.*)$/);
    }
#    print $outFolder,"\n";
    `mkdir -p $outFolder`; #Create the folder
    #Now first prune off the EDITED tags, and put them there
    my $prunedFile=$outFolder."/$filePrefix". ".". $suffix .".prunedparse";
    my $depsFile=$outFolder."/$filePrefix". ".". $suffix .".dep";
    #We make a sanity check too, and make sure they have the same number of sentences:
    my $origSenCount=countSentences($parse_file);
    my $origLineCount=countNonBlankLines($parse_file);
    print "Pruning off EDITED tags...";
    pruneEdits($parse_file,$prunedFile);
    print "Done.\n";
    my $prunedSenCount=countSentences($prunedFile);
    my $prunedLineCount=countNonBlankLines($prunedFile);
    my $cmd="java -cp $JAVA_CP clear.engine.PennToDep -l en -m $CLEAR_LEMMA_DICT -h $CLEAR_HEAD_RULES -f -i $prunedFile -o $depsFile";
    print "$cmd\n";
    `$cmd`;
    my $depsSenCount=countSentences($depsFile);
    my $depsLineCount=countNonBlankLines($depsFile);
    print "Done.\n";
    printf("%8s %10s %10s %10s\n","","Original","Pruned","Deps");
    printf("%-8s %10s %10s %10s\n","Sens",$origSenCount,$prunedSenCount,$depsSenCount);
    printf("%-8s %10s %10s %10s\n","Tokens",$origLineCount,$prunedLineCount,$depsLineCount);
    my $SUCCESS=($origSenCount==$prunedSenCount && $origSenCount==$depsSenCount && $prunedLineCount>=$depsLineCount && $depsLineCount<=$origLineCount);
    if($SUCCESS){
	print "Success\n";
    } else {
	print "Failed... aborting.\n";
	exit 1;
    }
    print "\n";
    
#    last if($fileCount==1);
}

print "\nConverted $fileCount files sucessfully.\n";

sub countSentences {
    my $file=shift;
    my $count=`egrep "^\\s*\$" $file | wc -l`;
    chomp $count;
    return $count;
}

sub countNonBlankLines {
    my $file=shift;
    my $count=`grep -v "^\\s*\$" $file | wc -l`;
    chomp $count;
    return $count;
}

sub pruneEdits {
    my $input=shift;
    my $output=shift;
    open IF,$input or die("Failed to open $input for reading");
    open OF,">$output" or die("Failed ot open $output for writing");
    my $tree;
    my $sencount=0;
    while(my $line=<IF>){
	if($line=~/^\s*$/) { #Whitespace only -- end of sentence
	    $sencount++;
	    if($tree=~/EDITED/){
#		print "--------------------\n";
#		print $tree,"\n";
		die("not wellformed input") if (!wellFormed($tree));
		my $x=$tree;
		$tree=~s/[ \t]*\(EDITED.*?(\((?:[^\(\)]+?|(?1))+?\))(?:\s*(\((?:[^\(\)]+?|(?2))+?\)))*\)/ /g;
		$tree=~s/\((S|NP)\s*\)//g;
		#and kill line breaks:
		my @lines=split(/\n/,$tree);
		@lines= grep(/\S/,@lines);
		$tree=join("\n",@lines)."\n";;
		if(!wellFormed($tree)){
		    print "====\n";
		    print "$x\n";
		    print "----\n";
		    print $tree,"\n";
		    print "----\n";
		    die("not wellformed output!") 
		}
	    }
	    print OF $tree,"\n";
	    $tree="";
	} else {
	    $tree.=$line;
	}
    }
    close IF;
    close OF;
}

sub wellFormed {
    my $tree=shift;
    my $left = () = $tree=~/\(/g;
    my $right = () = $tree=~/\)/g;
    if($left!=$right){
	print "Left: $left\n";
	print "Right: $right\n";
	return 0;
    }
    return 1;
}
